/*
 *	HT Editor
 *	htsearch.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "cstream.h"
#include "htapp.h"
#include "htatom.h"
#include "htctrl.h"
#include "hthist.h"
#include "htkeyb.h"
#include "htsearch.h"
#include "htstring.h"

extern "C" {
#include "evalx.h"
}

#include <stdlib.h>
#include <string.h>

/*
 *	CLASS ht_fxbin_search_request
 */
 
ht_fxbin_search_request::ht_fxbin_search_request(UINT search_class, UINT flags, UINT _data_size, byte *_data) :
	ht_search_request(search_class, ST_FXBIN, flags)
{
	data_size=_data_size;
	data=(byte*)malloc(data_size);
	memmove(data, _data, data_size);
}

ht_fxbin_search_request::~ht_fxbin_search_request()
{
	free(data);
}

object *ht_fxbin_search_request::duplicate()
{
	byte *dt=(byte*)malloc(data_size);
	memmove(dt, data, data_size);
	return new ht_fxbin_search_request(search_class, flags, data_size, dt);
}

/*
 *	CLASS ht_regex_search_request
 */
 
ht_regex_search_request::ht_regex_search_request(UINT search_class, UINT flags, char *regex) :
	ht_search_request(search_class, ST_REGEX, flags)
{
	rx_str=ht_strdup(regex);
	int r=regcomp(&rx, rx_str, REG_EXTENDED | ((flags & SF_REGEX_CASEINSENSITIVE) ? REG_ICASE : 0));
	if (r) throw new ht_regex_search_exception(r, &rx);
}

ht_regex_search_request::~ht_regex_search_request()
{
	regfree(&rx);
}

object *ht_regex_search_request::duplicate()
{
	return new ht_regex_search_request(search_class, flags, rx_str);
}

/*
 *	CLASS ht_regex_search_request
 */
 
ht_regex_search_exception::ht_regex_search_exception(int e, regex_t *r)
{
	errorcode=e;
	regex=r;
	
	char s[128];
	regerror(errorcode, regex, s, sizeof s-32);
	sprintf(rxerr, "error compiling regex: %s", s);
}
	
const char* ht_regex_search_exception::what()
{
	return rxerr;
}

/*
 *	CLASS ht_expr_search_request
 */

ht_expr_search_request::ht_expr_search_request(UINT search_class, UINT flags, char *Expr) :
	ht_search_request(search_class, ST_EXPR, flags)
{
	expr = ht_strdup(Expr);
}

ht_expr_search_request::~ht_expr_search_request()
{
	if (expr) delete expr;
}

object *ht_expr_search_request::duplicate()
{
	return new ht_expr_search_request(search_class, flags, expr);
}

/*
 *	binary search
 */

void bufdowncase(byte *buf, dword len)
{
	for (dword i=0; i<len; i++) {
		if ((buf[i]>='A') && (buf[i]<='Z')) buf[i]+=32;
	}
}

#include "htiobox.h"

#define SEARCH_BUF_SIZE 256*1024

int search_bin(ht_streamfile *file, FILEOFS start, FILEOFS end, dword pattern_len, byte *_pattern, FILEOFS *result_ofs, UINT flags, ht_progress_indicator *progress_indicator)
{
	int leave=0, success=0;

	byte *pattern=(byte*)malloc(sizeof *pattern * pattern_len);
	memmove(pattern, _pattern, sizeof *pattern * pattern_len);
	if (flags & SFBIN_CASEINSENSITIVE) bufdowncase(pattern, pattern_len);
	byte *buf=(byte*)malloc(SEARCH_BUF_SIZE);
	byte *bptr=buf;
	dword b;
	bool end_reached=0;
	FILEOFS o=start;
	FILEOFS next_progress_draw=o;
	dword next_progress_inc=(end-start)/100;

	dword pbyte=*pattern;
	file->seek(o);
	b=SEARCH_BUF_SIZE;
	if (b+o>end) b=end-o;

	dword filemode;

	file->cntl(FCNTL_CACHE_GET_MODE, &filemode);
	file->cntl(FCNTL_CACHE_SET_MODE, filemode & ~CM_READ_CACHED);

	int ggg2=new_timer();
	start_timer(ggg2);
	b=file->read(buf, b);
	stop_timer(ggg2);
//	LOG("debug: (search-read) %d msecs.", get_timer_msec(ggg2));
	delete_timer(ggg2);

	int ggg=new_timer();
	start_timer(ggg);
	if (flags & SFBIN_CASEINSENSITIVE) bufdowncase(buf, b);
	while (1) {
		bptr=buf;
		while ((bptr-buf+pattern_len<=b) && (bptr=(byte*)memchr(bptr, pbyte, b-pattern_len+1-(bptr-buf)))) {
			if (!memcmp(pattern, bptr, pattern_len)) {
				*result_ofs=o+bptr-buf;
				success=1;
				leave=1;
				break;
			}
			bptr++;
		}
/*		while (bptr-buf+pattern_len<=b) {
			if (pbyte==*bptr) {
				if (bufequal(pattern, bptr, pattern_len)) {
					*result_ofs=o+bptr-buf;
					success=1;
					leave=1;
					break;
				}
			}
			bptr++;
		}*/
		if (leave) break;
		if ((progress_indicator) && (o>=next_progress_draw)) {
			if (ht_keypressed()) {
				if (ht_getkey()==K_Escape) break;
			}
			
			char text[128];
			if ((end-start)/100) {
				int p=(o-start)/((end-start)/100);
				sprintf(text, "%d %%", p);
			} else {
				strcpy(text, "? %");
			}
			progress_indicator->settext(text);
			progress_indicator->sendmsg(msg_draw, 0);
			screen->show();
			next_progress_draw+=next_progress_inc;
		}
		o+=b-pattern_len;
		if (end_reached) break;
		file->seek(o);
		b=SEARCH_BUF_SIZE;
		if (b+o>=end) {
			b=end-o;
			end_reached=1;
		}
		b=file->read(buf, b);
		if (flags & SFBIN_CASEINSENSITIVE) bufdowncase(buf, b);
	}
	stop_timer(ggg);

	delete pattern;
	delete buf;

//	LOG("debug: (search) %d msecs.", get_timer_msec(ggg));
	delete_timer(ggg);

	file->cntl(FCNTL_CACHE_SET_MODE, filemode);

	return success;
}

/*
 *	CLASS ht_hexascii_search_mask
 */

void ht_hexascii_search_mask::init(bounds *b, int options, ht_list *history)
{
	ht_group::init(b, VO_SELECTABLE, "some hex/ascii search mask");
	VIEW_DEBUG_NAME("ht_hexascii_search_mask");

	bounds c;
/* ascii string */
	c.x=6;
	c.y=0;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* ascii string label */
	c.x=0;
	c.w=5;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "~ascii", str);
	insert(strlabel);

/* hex string */
	c.x=6;
	c.y=2;
	c.w=40;
	c.h=1;
	ht_hexinputfield *hex=new ht_hexinputfield();
	hex->init(&c, 64);
	hex->attach(str);
	insert(hex);
/* hex string label */
	c.x=0;
	c.w=5;
	c.h=1;
	ht_label *hexlabel=new ht_label();
	hexlabel->init(&c, "~hex", hex);
	insert(hexlabel);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from ofs", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to ofs", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_hexascii_search_mask::done()
{
	ht_group::done();
}

/*
 *	CLASS ht_evalstr_search_mask
 */

void ht_evalstr_search_mask::init(bounds *b, int options, ht_list *history)
{
	ht_group::init(b, VO_SELECTABLE, "some evalstr search mask");
	VIEW_DEBUG_NAME("ht_evalstr_search_mask");

	bounds c;
/* string */
	c.x=0;
	c.y=1;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* string label */
	c.x=0;
	c.y=0;
	c.w=12;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "Eval ~string", str);
	insert(strlabel);
/* hint */
/*	c.x=0;
	c.y=2;
	c.w=b->w-2;
	c.h=1;
	ht_statictext *hint=new ht_statictext();
	hint->init(&c, "(example: \"hello\\n\\0\\077\\xd\" 'ho',011b,66o)", 0);
	insert(hint);*/

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from ofs", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to ofs", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_evalstr_search_mask::done()
{
	ht_group::done();
}

/*
 *	CLASS ht_text_search_mask
 */

void ht_text_search_mask::init(bounds *b, int options, ht_list *history)
{
	ht_group::init(b, VO_SELECTABLE, "some text search mask");
	VIEW_DEBUG_NAME("ht_text_search_mask");

	bounds c;
/* string */
	c.x=0;
	c.y=1;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* string label */
	c.x=0;
	c.y=0;
	c.w=12;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "~regular expression", str);
	insert(strlabel);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from addr", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to addr", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_text_search_mask::done()
{
	ht_group::done();
}

/*
 *	CLASS ht_expr_search_mask
 */

void	ht_expr_search_mask::init(bounds *b, int options, ht_list *history=0)
{
	ht_group::init(b, VO_SELECTABLE, "some expression search mask");
	VIEW_DEBUG_NAME("ht_expr_search_mask");

	bounds c;
/* string */
	c.x=0;
	c.y=1;
	c.w=40;
	c.h=1;
	str=new ht_strinputfield();
	str->init(&c, 64, history);
	insert(str);
/* string label */
	c.x=0;
	c.y=0;
	c.w=12;
	c.h=1;
	ht_label *strlabel=new ht_label();
	strlabel->init(&c, "~Expression", str);
	insert(strlabel);
/* hint */
	c.x=0;
	c.y=2;
	c.w=b->w-2;
	c.h=1;
	ht_statictext *hint=new ht_statictext();
	hint->init(&c, "Search stops if expression evals to non-zero", 0);
	insert(hint);

/* range start */
	c.x=10;
	c.y=4;
	c.w=10;
	c.h=1;
	range_start=new ht_strinputfield();
	range_start->init(&c, 10);
	insert(range_start);
/* range start label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *rslabel=new ht_label();
	rslabel->init(&c, "~from ofs", range_start);
	insert(rslabel);

/* range end */
	c.x=10;
	c.y=6;
	c.w=10;
	c.h=1;
	range_end=new ht_strinputfield();
	range_end->init(&c, 10);
	insert(range_end);
/* range end label */
	c.x=0;
	c.w=9;
	c.h=1;
	ht_label *relabel=new ht_label();
	relabel->init(&c, "~to ofs", range_end);
	insert(relabel);
/* options */
	c.w=35;
	c.y=8;
	c.h=2;
	ht_string_list *opts=new ht_string_list();
	opts->init();
	opts->insert_string("~case insensitive");
//	opts->insert_string("find all occurencies (~idle)");
	option_boxes=new ht_checkboxes();
	option_boxes->init(&c, opts);
	ht_checkboxes_data d;
	d.state=options;
	option_boxes->databuf_set(&d);
	insert(option_boxes);
}

void ht_expr_search_mask::done()
{
	ht_group::done();
}

/*
 *	CLASS ht_expr_search_exception
 */
 
ht_expr_search_exception::ht_expr_search_exception(char *Error)
{
	error = ht_strdup(Error);
}

ht_expr_search_exception::~ht_expr_search_exception()
{
	if (error) delete error;
}

const char* ht_expr_search_exception::what()
{
	return error;
}

/*
 *
 */

void insert_search_history_entry(ht_clist *hist, ht_view *form, byte *str, UINT slen, byte *start, UINT startlen, byte *end, UINT endlen, UINT state)
{
	char s[1024], *tt=s;	/* FIXME: possible buffer overflow */
	tt+=escape_special(tt, sizeof s-(tt-s), str, slen, NULL, false);
	strcat(tt, " (");
	if (state & 1) {
		strcat(tt, "case-insensitive");
	} else {
		strcat(tt, "case-sensitive");
	}
	if (startlen) {
		strcat(tt, ", start=");
		tt+=strlen(tt);
		tt+=escape_special(tt, sizeof s-(tt-s), start, startlen);
	}
	if (endlen) {
		strcat(tt, ", end=");
		tt+=strlen(tt);
		tt+=escape_special(tt, sizeof s-(tt-s), end, endlen);
	}
	strcat(tt, ")");
	insert_history_entry(hist, s, form);
}

bool test_str_to_ofs(FILEOFS *ofs, byte *str, UINT strlen, ht_format_viewer *format, char *desc)
{
#define TEST_STR_TO_OFS_MAXSTRLEN		128
	if (strlen>TEST_STR_TO_OFS_MAXSTRLEN) {
		errorbox("%s: expression too long (len %d, max %d)", desc, strlen, TEST_STR_TO_OFS_MAXSTRLEN);
		return false;
	}
	if (strlen>0) {
		char s[TEST_STR_TO_OFS_MAXSTRLEN+1];
		bin2str(s, str, strlen);
		if (!format->string_to_offset(s, ofs)) {
			errorbox("%s: invalid expression: '%s'", desc, s);
			return false;
		}
	}
	return true;
}

bool test_str_to_addr(fmt_vaddress *addr, byte *str, UINT strlen, ht_format_viewer *format, char *desc)
{
#define TEST_STR_TO_ADDR_MAXSTRLEN		128
	if (strlen>TEST_STR_TO_ADDR_MAXSTRLEN) {
		errorbox("%s: expression too long (len %d, max %d)", desc, strlen, TEST_STR_TO_ADDR_MAXSTRLEN);
		return false;
	}
	if (strlen>0) {
		char s[TEST_STR_TO_ADDR_MAXSTRLEN+1];
		bin2str(s, str, strlen);
		if (!format->string_to_address(s, addr)) {
			errorbox("%s: invalid expression: '%s'", desc, s);
			return false;
		}
	}
	return true;
}

/* FIXME: global var */
UINT lastsearchmodeid=0;

void search_dialog(ht_format_viewer *format, UINT searchmodes)
{
	if (!searchmodes) return;
	
	bounds b;
	b.w=50;
	b.h=15;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	ht_search_dialog *dialog=new ht_search_dialog();
	dialog->init(&b);
	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;

	int dofs=0;

	ht_clist *hist_hexascii=0;
	if (searchmodes & SEARCHMODE_BIN) {
		ht_hexascii_search_mask *s0=new ht_hexascii_search_mask();
		hist_hexascii=(ht_clist*)find_atom(HISTATOM_SEARCH_BIN);
		s0->init(&b, 0, hist_hexascii);
		dialog->insertmode(0, "bin: ascii/hex", s0);
		dofs+=dialog->datasize();
	}

	ht_clist *hist_evalstr=0;
	if (searchmodes & SEARCHMODE_EVALSTR) {
		ht_evalstr_search_mask *s1=new ht_evalstr_search_mask();
		hist_evalstr=(ht_clist*)find_atom(HISTATOM_SEARCH_EVALSTR);
		s1->init(&b, 0, hist_evalstr);
		dialog->insertmode(1, "bin: eval str", s1);
		dofs+=dialog->datasize();
	}

	ht_clist *hist_text=0;
	if (searchmodes & SEARCHMODE_TEXT) {
		ht_text_search_mask *s1=new ht_text_search_mask();
		hist_text=(ht_clist*)find_atom(HISTATOM_SEARCH_TEXT);
		s1->init(&b, 0, hist_text);
		dialog->insertmode(2, "display: regex", s1);
		dofs+=dialog->datasize();
	}

	ht_clist *hist_expr=0;
	if (searchmodes & SEARCHMODE_EXPR) {
		ht_expr_search_mask *s1=new ht_expr_search_mask();
		hist_expr=(ht_clist*)find_atom(HISTATOM_SEARCH_EXPR);
		s1->init(&b, 0, hist_expr);
		dialog->insertmode(3, "expr non-zero", s1);
		dofs+=dialog->datasize();
	}

	dialog->selectmode(lastsearchmodeid);

	if (dialog->run(0)) {
		ht_search_dialog_data *t=(ht_search_dialog_data *)malloc(dialog->datasize());
		dialog->databuf_get(t);

		int modeid=t->modes_id;

		free(t);

		ht_view *form=dialog->getmodeform();
		UINT ds=form->datasize();
		void *formdata=malloc(ds);
		form->databuf_get(formdata);
		
		lastsearchmodeid=modeid;
		
		switch (modeid) {
			case 0: {
/* ascii/hex */
				ht_hexascii_search_mask_data *d=(ht_hexascii_search_mask_data*)formdata;
				if (d->str.textlen) {
					insert_search_history_entry(hist_hexascii, form, d->str.text, d->str.textlen, d->start.text, d->start.textlen, d->end.text, d->end.textlen, d->options.state);

					FILEOFS start=0, end=0xffffffff;
					format->get_current_offset(&start);

					if (test_str_to_ofs(&start, d->start.text, d->start.textlen, format, "start-offset")
					&& test_str_to_ofs(&end, d->end.text, d->end.textlen, format, "end-offset")) {
//						infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
						ht_fxbin_search_request *q=new ht_fxbin_search_request(SC_PHYSICAL, d->options.state & 1 ? SF_FXBIN_CASEINSENSITIVE: 0, d->str.textlen, d->str.text);
						ht_physical_search_result *r=(ht_physical_search_result*)format->psearch(q, start, end);
						if (r) {
							format->goto_offset(r->offset, format);
							format->pselect_set(r->offset, r->offset+r->size);
							delete r;
						} else infobox("string not found");
						delete q;
					}
				}
				break;
			}
			case 1: {
/* EVAL STR */
#define EVALSTR_MAXSTRLEN		128
				ht_evalstr_search_mask_data *d=(ht_evalstr_search_mask_data*)formdata;
				if (d->str.textlen) {
					insert_search_history_entry(hist_evalstr, form, d->str.text, d->str.textlen, d->start.text, d->start.textlen, d->end.text, d->end.textlen, d->options.state);

					char strbuf[EVALSTR_MAXSTRLEN+1];
					if (d->str.textlen<=EVALSTR_MAXSTRLEN) {
						bin2str(strbuf, d->str.text, d->str.textlen);

						scalar_t r;
						str_t s, *t=NULL;
						if (eval(&r, strbuf, NULL, NULL, NULL)) {
							t=&s;
							scalar_context_str(&r, t);
							scalar_destroy(&r);
						}

						FILEOFS start=0, end=0xffffffff;
						format->get_current_offset(&start);

						if (t->value && t->len) {
							if (test_str_to_ofs(&start, d->start.text, d->start.textlen, format, "start-offset")
							&& test_str_to_ofs(&end, d->end.text, d->end.textlen, format, "end-offset")) {
//								infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
								ht_fxbin_search_request *q=new ht_fxbin_search_request(SC_PHYSICAL, d->options.state & 1 ? SF_FXBIN_CASEINSENSITIVE: 0, t->len, (byte*)t->value);
								ht_physical_search_result *r=(ht_physical_search_result*)format->psearch(q, start, end);
								if (r) {
									format->goto_offset(r->offset, format);
									format->pselect_set(r->offset, r->offset+r->size);
									delete r;
								} else infobox("string not found");
								delete q;
							}
						} else {
							char *str;
							int pos;
							get_eval_error(&str, &pos);
							errorbox("eval error at pos %d: %s", pos, str);
						}
					} else errorbox("expression too long (size %d, max %d)", strlen, EVALSTR_MAXSTRLEN);
				}
				break;
			}
			case 2: {
/* text */
#define TEXT_MAXSTRLEN		128
				ht_text_search_mask_data *d=(ht_text_search_mask_data*)formdata;
				if (d->str.textlen) {
					insert_search_history_entry(hist_text, form, d->str.text, d->str.textlen, d->start.text, d->start.textlen, d->end.text, d->end.textlen, d->options.state);

					char strbuf[TEXT_MAXSTRLEN+1];
					if (d->str.textlen<=TEXT_MAXSTRLEN) {
						bin2str(strbuf, d->str.text, d->str.textlen);

						fmt_vaddress start=0, end=0xffffffff;
						format->get_current_address(&start);

						if (test_str_to_addr(&start, d->start.text, d->start.textlen, format, "start-address")
						&& test_str_to_addr(&end, d->end.text, d->end.textlen, format, "end-address")) {
//							infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
							try {
								ht_regex_search_request *q=new ht_regex_search_request(SC_VISUAL, d->options.state & 1 ? SF_REGEX_CASEINSENSITIVE : 0, strbuf);
								ht_visual_search_result *r=(ht_visual_search_result *)format->vsearch(q, start, end);
								if (r) {
									format->goto_address(r->address, format);
									delete r;
								} else infobox("string not found");
								delete q;
							} catch (ht_exception *e) {
								errorbox("error: %s", e->what());
							}
						}
					} else errorbox("expression too long (size %d, max %d)", strlen, TEXT_MAXSTRLEN);
				}
				break;
			}
			case 3: {
/* expression */
#define EXPR_MAXSTRLEN		256
				ht_expr_search_mask_data *d=(ht_expr_search_mask_data*)formdata;
				if (d->str.textlen) {
					insert_search_history_entry(hist_expr, form, d->str.text, d->str.textlen, d->start.text, d->start.textlen, d->end.text, d->end.textlen, d->options.state);

					char strbuf[EXPR_MAXSTRLEN+1];
					if (d->str.textlen<=EXPR_MAXSTRLEN) {
						bin2str(strbuf, d->str.text, d->str.textlen);

						FILEOFS start=0, end=0xffffffff;
						format->get_current_offset(&start);

						if (test_str_to_ofs(&start, d->start.text, d->start.textlen, format, "start-offset")
						&& test_str_to_ofs(&end, d->end.text, d->end.textlen, format, "end-offset")) {
//							infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
							try {
								ht_expr_search_request *q=new ht_expr_search_request(SC_PHYSICAL, 0, strbuf);
								ht_physical_search_result *r=(ht_physical_search_result *)format->psearch(q, start, end);
								if (r) {
									format->goto_offset(r->offset, format);
//									format->pselect_set(r->offset, r->offset+r->size);
									delete r;
								} else infobox("no match");
								delete q;
							} catch (ht_exception *e) {
								errorbox("error: %s", e->what());
							}
						}
					} else errorbox("expression too long (size %d, max %d)", strlen, TEXT_MAXSTRLEN);
				}
				break;
			}
			default:
				warnbox("search mode (modeid=%d) not supported !", modeid);
		}
		free(formdata);
	}
	dialog->done();
	delete dialog;
}


